from collections import defaultdict
from typing import Generator, List

from .entry import (
    RequirementsConstraintsEntry,
    RequirementsEditableEntry,
    RequirementsEntry,
    RequirementsIndexURLEntry,
    RequirementsPackageEntry,
    RequirementsRecursiveEntry,
    RequirementsVCSPackageEntry,
    RequirementsWheelPackageEntry,
)
from .error import ConstraintWithoutNameError
from .identify_package_list_file_type import (
    PackageListFileType,
    identify_package_list_file_type,
)
from .parse_requirements_txt import parse_requirements_txt
from .parse_setup_py import parse_setup_py


def _list_packages_from_files(
    file_paths: List[str],
    *,
    recurse_recursive: bool = False,
    recurse_editable: bool = False,
    remove_editable: bool = False,
    remove_recursive: bool = False,
    remove_constraints: bool = False,
    remove_vcs: bool = False,
    remove_wheel: bool = False,
    remove_unversioned: bool = False,
    remove_index_urls: bool = False,
) -> Generator[RequirementsEntry, None, None]:
    generators = []

    for file_path in file_paths:
        package_list_file_type = identify_package_list_file_type(file_path)
        if package_list_file_type == PackageListFileType.REQUIREMENTS_TXT:
            generators.append(parse_requirements_txt(file_path))
        elif package_list_file_type == PackageListFileType.SETUP_PY:
            generators.append(parse_setup_py(file_path))

    while len(generators) > 0:
        for requirement in generators[0]:
            if isinstance(requirement, RequirementsRecursiveEntry):
                if recurse_recursive:
                    generators.append(
                        parse_requirements_txt(requirement.absolute_path)
                    )
                elif not remove_recursive:
                    yield requirement
            elif isinstance(requirement, RequirementsConstraintsEntry):
                if not remove_constraints:
                    yield requirement
            elif isinstance(requirement, RequirementsEditableEntry):
                if recurse_editable:
                    generators.append(
                        parse_setup_py(
                            requirement.resolved_absolute_path,
                            requirement.extras,
                        )
                    )
                elif not remove_editable:
                    yield requirement
            elif isinstance(requirement, RequirementsIndexURLEntry):
                if not remove_index_urls:
                    yield requirement
            elif isinstance(requirement, RequirementsVCSPackageEntry):
                if not remove_vcs:
                    yield requirement
            elif isinstance(requirement, RequirementsWheelPackageEntry):
                if not remove_wheel:
                    yield requirement
            elif isinstance(requirement, RequirementsPackageEntry):
                if remove_unversioned and not requirement.version:
                    continue
                else:
                    yield requirement

        generators = generators[1:]


def _dedupe_requirements(
    generator: Generator[RequirementsEntry, None, None]
) -> Generator[RequirementsEntry, None, None]:
    """Removes exact duplicates from the list of requirements.

    Duplicates can happen as multiple files are merged together and they
    all reefer to the same dependency. De-duping should have no impact
    on the final result, it just makes the list easier to browse.
    """

    unique_requirements = []
    unique_requirements_strings = set()
    for requirement in generator:
        if str(requirement) in unique_requirements_strings:
            continue

        unique_requirements.append(requirement)
        unique_requirements_strings.add(str(requirement))

    for requirement in unique_requirements:
        yield requirement


def _inline_constraints(
    generator: Generator[RequirementsEntry, None, None]
) -> Generator[RequirementsEntry, None, None]:
    """Inlines constraints specified in constraint.txt files (specified by -c).

    Constraints override what version of a package should be installed. They cannot
    add new packages to the list. Any package in the constraints file that isn't
    already in the list is ignored.

    We have to be careful because the constraints file could contain multiple
    entries for the same package. Each entry could have different markers.

    Constraint files have the same syntax as requirements.txt files, but with
    some restrictions:

        * No nesting
        * No editables
        * All entries must have a name
    """

    requirements = []
    constraints = defaultdict(list)

    for requirement in generator:
        if not isinstance(requirement, RequirementsConstraintsEntry):
            requirements.append(requirement)
            continue

        for constraint in parse_requirements_txt(requirement.absolute_path):
            if constraint.package_name() is None:
                raise ConstraintWithoutNameError(constraint)

            constraints[constraint.package_name()].append(constraint)

    for requirement in requirements:
        if not requirement.package_name():
            yield requirement
            continue

        requirement_constraints = constraints.get(requirement.package_name())
        if requirement_constraints:
            for constraint in requirement_constraints:
                yield constraint
        else:
            yield requirement


def list_packages_from_files(
    file_paths: List[str],
    *,
    recurse_recursive: bool = False,
    recurse_editable: bool = False,
    inline_constraints: bool = False,
    remove_editable: bool = False,
    remove_recursive: bool = False,
    remove_constraints: bool = False,
    remove_vcs: bool = False,
    remove_wheel: bool = False,
    remove_unversioned: bool = False,
    remove_index_urls: bool = False,
    dedupe: bool = False,
) -> Generator[RequirementsEntry, None, None]:
    generator = _list_packages_from_files(
        file_paths,
        recurse_recursive=recurse_recursive,
        recurse_editable=recurse_editable,
        remove_editable=remove_editable,
        remove_recursive=remove_recursive,
        remove_constraints=remove_constraints,
        remove_vcs=remove_vcs,
        remove_wheel=remove_wheel,
        remove_unversioned=remove_unversioned,
        remove_index_urls=remove_index_urls,
    )

    if inline_constraints:
        generator = _inline_constraints(generator)

    if dedupe:
        generator = _dedupe_requirements(generator)

    for requirement in generator:
        yield requirement



import os
import re

from typing import Generator, List, Optional, Tuple
from urllib.parse import urlparse

from .entry import (
    RequirementsConstraintsEntry,
    RequirementsDirectRefEntry,
    RequirementsEditableEntry,
    RequirementsEntry,
    RequirementsEntrySource,
    RequirementsIndexURLEntry,
    RequirementsPackageEntry,
    RequirementsRecursiveEntry,
    RequirementsVCSPackageEntry,
    RequirementsWheelPackageEntry,
)

# From: https://pip.pypa.io/en/stable/topics/vcs-support/#supported-vcs
vcs_protocols = [
    "git+file",
    "git+https",
    "git+ssh",
    "git+http",
    "git+git",
    "git+",
    "hg+",
    "hg+file",
    "hg+http",
    "hg+https",
    "hg+ssh",
    "hg+static-http",
    "svn+",
    "svn+svn",
    "svn+http",
    "svn+https",
    "svn+ssh",
    "bzr+http",
    "bzr+https",
    "bzr+ssh",
    "bzr+sftp",
    "bzr+ftp",
    "bzr+lp",
]
vcs_protocols_regex = r"^(%s)" % "|".join(
    [re.escape(protocol) for protocol in vcs_protocols]
)


class RequirementsEntryParseError(RuntimeError):
    pass


def parse_requirements_list(
    source: Optional[RequirementsEntrySource], lines: List[str]
) -> Generator[RequirementsEntry, None, None]:

    for index, line in enumerate(lines):
        stripped_line = _clean_line(line)

        if not len(stripped_line) or stripped_line.startswith("#"):
            continue

        line_source = None
        if source:
            line_source = RequirementsEntrySource(
                path=source.path, line=stripped_line, line_number=index + 1
            )

        requirement, extras, markers = _split_requirement_and_markers(
            stripped_line
        )

        if requirement.startswith("-r"):
            yield parse_recursive_requirements_entry(
                line_source, requirement, extras, markers
            )

        elif requirement.startswith("-c"):
            yield parse_constraints_requirements_entry(
                line_source, requirement, extras, markers
            )

        elif requirement.startswith("-e"):
            yield parse_editable_requirements_entry(
                line_source, requirement, extras, markers
            )

        elif requirement.startswith("-i"):
            yield parse_index_url_requirements_entry(
                line_source, requirement, extras, markers
            )

        elif re.match(vcs_protocols_regex, requirement):
            yield parse_vcs_requirements_entry(
                line_source, requirement, extras, markers
            )

        elif "@" in requirement:
            yield parse_direct_ref_requirements_entry(
                line_source, requirement, extras, markers,
            )

        elif re.match(r"^(http|https|file)://(.+).whl", requirement):
            yield parse_wheel_requirements_entry(
                line_source, requirement, extras, markers
            )

        else:
            yield parse_package_requirements_entry(
                line_source, requirement, extras, markers
            )


def parse_recursive_requirements_entry(
    source: RequirementsEntrySource,
    requirement: str,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsRecursiveEntry:
    original_path = _clean_line(requirement.lstrip("-r"))
    absolute_path = os.path.realpath(
        os.path.join(os.path.dirname(source.path), original_path)
    )

    return RequirementsRecursiveEntry(
        source=source, original_path=original_path, absolute_path=absolute_path,
    )


def parse_constraints_requirements_entry(
    source: RequirementsEntrySource,
    requirement: str,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsConstraintsEntry:
    original_path = _clean_line(requirement.lstrip("-c"))
    absolute_path = os.path.realpath(
        os.path.join(os.path.dirname(source.path), original_path)
    )

    return RequirementsConstraintsEntry(
        source=source, original_path=original_path, absolute_path=absolute_path,
    )


def parse_editable_requirements_entry(
    source: RequirementsEntrySource,
    requirement: str,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsEditableEntry:
    original_path = _clean_line(requirement.lstrip("-e"))
    resolved_path = original_path

    if not original_path.endswith(".py"):
        resolved_path = os.path.join(original_path, "setup.py")

    absolute_path = os.path.realpath(
        os.path.join(os.path.dirname(source.path), original_path)
    )
    resolved_absolute_path = os.path.realpath(
        os.path.join(os.path.dirname(source.path), resolved_path)
    )

    return RequirementsEditableEntry(
        source=source,
        original_path=original_path,
        absolute_path=absolute_path,
        resolved_path=resolved_path,
        resolved_absolute_path=resolved_absolute_path,
        extras=extras,
    )


def parse_index_url_requirements_entry(
    source: RequirementsEntrySource,
    requirement: str,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsIndexURLEntry:
    url = _clean_line(requirement.lstrip("-i"))
    return RequirementsIndexURLEntry(source=source, url=url)


def parse_vcs_requirements_entry(
    source: RequirementsEntrySource,
    requirement: str,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsVCSPackageEntry:
    vcs, uri = requirement.split("+", maxsplit=1)

    # If this looks like a normal URL, let's use urlparse. We need to do this
    # so we can tell the username and tag apart, which both use '@'
    match = re.match(r"^(.+)://", uri)
    if match:
        uri_parts = urlparse(uri)

        name = None
        if "egg=" in uri_parts.fragment:
            name = uri_parts.fragment.replace("egg=", "")
            uri_parts = uri_parts._replace(
                fragment=uri_parts.fragment.replace(f"egg={name}", "")
            )

        tag = None
        if "@" in uri_parts.path:
            path, tag = uri_parts.path.split("@", maxsplit=1)
            uri_parts = uri_parts._replace(path=path)

        uri = uri_parts.geturl()

        return RequirementsVCSPackageEntry(
            source=source, vcs=vcs, uri=uri, tag=tag, name=name
        )

    # We're dealing with a Git SSH uri: git@github.com:/repo.git@tag@egg=name
    name = None
    if "#egg=" in uri:
        uri, name = uri.split("#egg=", maxsplit=1)

    tag = None

    repo_name_index = uri.rfind(":")
    if repo_name_index >= 0:
        repo_name = uri[repo_name_index:]
        if "@" in repo_name:
            uri, tag = uri.rsplit("@", maxsplit=1)

    return RequirementsVCSPackageEntry(
        source=source, vcs=vcs, uri=uri, tag=tag, name=name
    )


def parse_wheel_requirements_entry(
    source: RequirementsEntrySource,
    requirement,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsWheelPackageEntry:
    uri = requirement
    name = None

    if "#egg=" in requirement:
        uri, name = requirement.split("#egg=")

    return RequirementsWheelPackageEntry(
        source=source, uri=uri, name=name, markers=markers
    )


def parse_direct_ref_requirements_entry(
    source: RequirementsEntrySource,
    requirement,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsDirectRefEntry:
    name, uri = requirement.split("@")
    name = name.strip()
    uri = uri.strip()

    return RequirementsDirectRefEntry(
        source=source, name=name, uri=uri, markers=markers,
    )


def parse_package_requirements_entry(
    source: RequirementsEntrySource,
    requirement: str,
    extras: List[str],
    markers: Optional[str] = None,
) -> RequirementsPackageEntry:
    operators = ["==", ">=", ">", "<=", "<"]
    for operator in operators:
        parts = requirement.split(operator)
        if len(parts) == 1:
            continue

        return RequirementsPackageEntry(
            source=source,
            name=parts[0],
            extras=extras,
            operator=operator,
            version=parts[1],
            markers=markers,
        )

    return RequirementsPackageEntry(
        source=source, name=requirement, extras=extras, markers=markers,
    )


def _clean_line(line: str) -> str:
    return (
        line.strip()
        .replace("\n", "")
        .replace("\r", "")
        .replace("  ", " ")
        .replace("  ", " ")
    )


def _split_requirement_and_markers(
    line: str,
) -> Tuple[str, List[str], Optional[str]]:
    requirement = line
    markers = None
    extras = []

    if ";" in requirement:
        requirement, markers = requirement.split(";")

        requirement = requirement.strip()
        markers = markers.strip()

    extras_match = re.search(re.compile(r"\[.*\]"), requirement)
    if extras_match:
        matched_text = extras_match[0]

        requirement = requirement.replace(matched_text, "")
        extras = [
            extra.strip()
            for extra in matched_text.strip("[").strip("]").strip().split(",")
        ]

    return requirement, extras, markers



from .entry import RequirementsEntry


class ConstraintWithoutNameError(RuntimeError):
    def __init__(self, requirement: RequirementsEntry) -> None:
        super().__init__(f"Constraint '{requirement}' does not have a name")



import argparse
import sys

from .list_packages_from_files import list_packages_from_files


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--recurse-recursive",
        default=False,
        help="recurse into -r entries",
        action="store_true",
    )
    parser.add_argument(
        "--recurse-editable",
        default=False,
        help="recurse into -e entries",
        action="store_true",
    )
    parser.add_argument(
        "--inline-constraints",
        default=False,
        help="recurse into -c entries and inline them",
        action="store_true",
    )
    parser.add_argument(
        "--dedupe",
        default=False,
        help="de-duplicate the resulting list",
        action="store_true",
    )
    parser.add_argument(
        "--remove-editable",
        default=False,
        help="remove editable requirements from the final list",
        action="store_true",
    )
    parser.add_argument(
        "--remove-recursive",
        default=False,
        help="remove recursive requirements (-r) from the final list",
        action="store_true",
    )
    parser.add_argument(
        "--remove-constraints",
        default=False,
        help="remove constaints (-c) from the final list",
        action="store_true",
    )
    parser.add_argument(
        "--remove-vcs",
        default=False,
        help="remove vcs requirements from the final list",
        action="store_true",
    )
    parser.add_argument(
        "--remove-wheel",
        default=False,
        help="remove wheel requirements from the final list",
        action="store_true",
    )
    parser.add_argument(
        "--remove-unversioned",
        default=False,
        help="remove requirements without a version number from the final list",
        action="store_true",
    )
    parser.add_argument(
        "--remove-index-urls",
        default=False,
        help="remove -i entries (index urls) from the final list",
        action="store_true",
    )
    parser.add_argument(
        "file_paths",
        nargs="+",
        help="list of requirements.txt or setup.py files",
    )

    args = parser.parse_args()

    requirements = list_packages_from_files(
        args.file_paths,
        recurse_recursive=args.recurse_recursive,
        recurse_editable=args.recurse_editable,
        inline_constraints=args.inline_constraints,
        remove_editable=args.remove_editable,
        remove_recursive=args.remove_recursive,
        remove_constraints=args.remove_constraints,
        remove_vcs=args.remove_vcs,
        remove_wheel=args.remove_wheel,
        remove_unversioned=args.remove_unversioned,
        remove_index_urls=args.remove_index_urls,
        dedupe=args.dedupe,
    )

    for req in requirements:
        print(req)

    return 0


if __name__ == "__main__":
    sys.exit(main())
